"""Ecos client custom exceptions."""


class EcosApiError(Exception):
    """Base exception class for all ECOS API-related errors."""


class InvalidJsonError(EcosApiError):
    """Raised when the API returns invalid JSON."""

    def __init__(self) -> None:
        super().__init__("Invalid JSON")


class ApiResponseError(EcosApiError):
    """Raised when the API returns a non-successful response."""

    def __init__(self, code: int, message: str) -> None:
        self.code = code
        self.message = message
        super().__init__(f"API call failed: {code} {message}")


class HttpError(EcosApiError):
    """Raised when an HTTP error occurs while making an API request."""

    def __init__(self, status_code: int, message: str) -> None:
        self.status_code = status_code
        self.message = message
        super().__init__(f"HTTP error: {status_code} {message}")



"""Implementation of a class for interacting with the ECOS API."""

from datetime import datetime
import logging
import time
from typing import Any

import requests

from ._base import JSON, _BaseEcos
from ._exceptions import ApiResponseError, HttpError, InvalidJsonError

# Configure logging
logger = logging.getLogger(__name__)


class Ecos(_BaseEcos):
    """Synchronous ECOS API client class.

    This class provides methods for interacting with the ECOS API, including
    authentication, retrieving user information, and managing homes. It uses
    the `requests` library to make HTTP requests to the API.
    """

    def _get(self, api_path: str, payload: dict[str, Any] = {}) -> JSON:
        """Make a GET request to the ECOS API.

        Args:
            api_path (str): The path of the API endpoint.
            payload (dict): The data to be sent with the request.

        Returns:
            JSON: The data returned by the API.

        Raises:
            requests.exceptions.HTTPError: If the API returns an HTTP error.
            ValueError: If the API returns a non-successful response.

        """
        api_path = api_path.lstrip("/")  # remove / from beginning of api_path
        full_url = self.url + "/" + api_path
        logger.info("API GET call: %s", full_url)
        try:
            response = requests.get(
                full_url, params=payload, headers={"Authorization": self.access_token}
            )
            logger.debug(response.text)
            body = response.json()
        except requests.exceptions.JSONDecodeError as err:
            if response.status_code != 200:
                raise HttpError(response.status_code, response.text) from err
            raise InvalidJsonError from err
        else:
            if not response.ok:
                raise HttpError(response.status_code, body["message"])
            if not body["success"]:
                logger.debug(body)
                raise ApiResponseError(body["code"], body["message"])
        return body["data"]

    def _post(self, api_path: str, payload: JSON = {}) -> JSON:
        """Make a POST request to the ECOS API.

        Args:
            api_path (str): The path of the API endpoint.
            payload (JSON): The data to be sent with the request.

        Returns:
            JSON: The data returned by the API.

        Raises:
            requests.exceptions.HTTPError: If the API returns an HTTP error.
            ValueError: If the API returns a non-successful response.

        """
        api_path = api_path.lstrip("/")  # remove / from beginning of api_path
        full_url = self.url + "/" + api_path
        logger.info("API POST call: %s", full_url)
        try:
            response = requests.post(
                full_url, json=payload, headers={"Authorization": self.access_token}
            )
            logger.debug(response.text)
            body = response.json()
        except requests.exceptions.JSONDecodeError as err:
            if response.status_code != 200:
                raise HttpError(response.status_code, response.text) from err
            raise InvalidJsonError from err
        else:
            if not response.ok:
                raise HttpError(response.status_code, body["message"])
            if not body["success"]:
                logger.debug(body)
                raise ApiResponseError(body["code"], body["message"])
        return body["data"]

    def login(self, email: str, password: str) -> None:
        """Authenticate with the ECOS API using a provided email and password.

        Args:
            email (str): The user's email to use for authentication.
            password (str): The user's password to use for authentication.

        """
        logger.info("Login")
        payload = {
            "_t": int(time.time()),
            "clientType": "BROWSER",
            "clientVersion": "1.0",
            "email": email,
            "password": password,
        }
        data = self._post("/api/client/guide/login", payload=payload)
        self.access_token = data["accessToken"]
        self.refresh_token = data["refreshToken"]

    def get_user_info(self) -> JSON:
        """Get user details.

        Returns:
            Details concerning the user. Example:
                ``` py
                {
                    "username": "john.doe@acme.com",
                    "nickname": "JohnD",
                    "email": "john.doe@acme.com",
                    "phone": "",
                    "timeZoneId": "209",
                    "timeZone": "GMT-05:00",
                    "timezoneName": "America/Toronto",
                    "datacenterPhoneCode": 49,
                    "datacenter": "EU",
                    "datacenterHost": "https://api-ecos-eu.weiheng-tech.com"
                }
                ```

        """
        logger.info("Get user info")
        return self._get("/api/client/settings/user/info")

    def get_homes(self) -> JSON:
        """Get a list of homes.

        Returns:
            A list of homes. Example:
                ``` py
                [
                    {
                        "homeId": "1234567890123456789",
                        "homeName": "SHARED_DEVICES",
                        "homeType": 0,
                        "longitude": None,
                        "latitude": None,
                        "homeDeviceNumber": 1,
                        "relationType": 1,
                        "createTime": 946684800000,
                        "updateTime": 946684800000,
                    },
                    {
                        "homeId": "9876543210987654321",
                        "homeName": "My Home",
                        "homeType": 1,
                        "longitude": None,
                        "latitude": None,
                        "homeDeviceNumber": 0,
                        "relationType": 1,
                        "createTime": 946684800000,
                        "updateTime": 946684800000,
                    },
                ]
                ```

        """
        logger.info("Get home list")
        home_list: list[Any] = self._get("/api/client/v2/home/family/query")
        for (
            home
        ) in home_list:  # force the name of the home for shared devices (homeType=0)
            if home["homeType"] == "0":
                home["homeName"] = "SHARED_DEVICES"
        return home_list

    def get_devices(self, home_id: int) -> JSON:
        """Get a list of devices for a home.

        Args:
            home_id (int): The home ID to get devices for.

        Returns:
            A list of devices. Example:
                ``` py
                [
                    {
                        "deviceId": "1234567890123456789",
                        "deviceAliasName": "My Device",
                        "state": 0,
                        "batterySoc": 0.0,
                        "batteryPower": 0,
                        "socketSwitch": None,
                        "chargeStationMode": None,
                        "vpp": False,
                        "type": 1,
                        "deviceSn": "SHC000000000000001",
                        "agentId": "9876543210987654321",
                        "lon": 0.0,
                        "lat": 0.0,
                        "deviceType": "XX-XXX123       ",
                        "resourceSeriesId": 101,
                        "resourceTypeId": 7,
                        "master": 0,
                        "emsSoftwareVersion": "000-00000-00",
                        "dsp1SoftwareVersion": "111-11111-11",
                    },
                ]
                ```

        """
        logger.info("Get devices for home %d", home_id)
        return self._get(
            "/api/client/v2/home/device/query", payload={"homeId": home_id}
        )

    def get_all_devices(self) -> JSON:
        """Get a list of all the devices.

        Returns:
            A list of devices. Example:
                ``` py
                [
                    {
                        "deviceId": "1234567890123456789",
                        "deviceAliasName": "My Device",
                        "wifiSn": "azerty123456789azertyu",
                        "state": 0,
                        "weight": 0,
                        "temp": None,
                        "icon": None,
                        "vpp": False,
                        "master": 0,
                        "type": 1,
                        "deviceSn": "SHC000000000000001",
                        "agentId": "",
                        "lon": 0.0,
                        "lat": 0.0,
                        "category": None,
                        "model": None,
                        "deviceType": None,
                    },
                ]
                ```

        """
        logger.info("Get devices for every homes")
        return self._get("/api/client/home/device/list")

    def get_realtime_device_data(self, device_id: int) -> JSON:
        """Get power metrics of the current day until now.

        Args:
            device_id (int): The device ID to get power metrics for.

        Returns:
            Multiple metrics of the current day. Example:
                ``` py
                {
                    "solarPowerDps": {
                        "946685100": 0.0,
                        "946685400": 0.0,
                        ...
                        "946733700": 0.0,
                    },
                    "batteryPowerDps": {...},
                    "gridPowerDps": {...},
                    "meterPowerDps": {...},
                    "homePowerDps": {...},
                    "epsPowerDps": {...},
                }
                ```

        """
        logger.info("Get current day data for device %d", device_id)
        return self._post(
            "/api/client/home/now/device/realtime", payload={"deviceId": device_id}
        )

    def get_realtime_home_data(self, home_id: int) -> JSON:
        """Get current power for the home.

        Args:
            home_id (int): The home ID to get current power for.

        Returns:
            Power data. Example:
                ``` py
                {
                    "batteryPower": 0,
                    "epsPower": 0,
                    "gridPower": 23,
                    "homePower": 1118,
                    "meterPower": 1118,
                    "solarPower": 0,
                    "chargePower": 0,
                    "batterySocList": [
                        {
                            "deviceSn": "SHC000000000000001",
                            "batterySoc": 0.0,
                            "sysRunMode": 1,
                            "isExistSolar": True,
                            "sysPowerConfig": 3,
                        }
                    ],
                }
                ```

        """
        logger.info("Get realtime data for home %d", home_id)
        return self._get(
            "/api/client/v2/home/device/runData", payload={"homeId": home_id}
        )

    def get_history(
        self, device_id: int, start_date: datetime, period_type: int
    ) -> JSON:
        """Get aggregated energy for a period.

        Args:
            device_id (int): The device ID to get history for.
            start_date (datetime): The start date.
            period_type (int): Possible value:

                - `0`: daily values of the calendar month corresponding to `start_date`
                - `1`: today daily values (`start_date` is ignored) (?)
                - `2`: daily values of the current month (`start_date` is ignored)
                - `3`: same than 2 ?
                - `4`: total for the current month (`start_date` is ignored)

        Returns:
            Data and metrics corresponding to the defined period. Example:
                ``` py
                {
                    "energyConsumption": 1221.2,
                    "solarPercent": 47.0,
                    "homeEnergyDps": {
                        "1733112000": 39.6,
                        "1733198400": 68.1,
                        "1733284800": 75.3,
                        ...
                        "1735707599": 41.3,
                    },
                }
                ```

        """
        logger.info("Get history for device %d", device_id)
        start_ts = int(start_date.timestamp())
        return self._post(
            "/api/client/home/history/home",
            payload={
                "deviceId": device_id,
                "timestamp": start_ts,
                "periodType": period_type,
            },
        )

    def get_insight(
        self, device_id: int, start_date: datetime, period_type: int
    ) -> JSON:
        """Get energy metrics and statistics of a device for a period.

        Args:
            device_id (int): The device ID to get data for.
            start_date (datetime): The start date.
            period_type (int): Possible value:

                - `0`: 5-minute power measurement for the calendar day corresponding to `start_date` (`insightConsumptionDataDto` is `None`)
                - `1`: (not implemented)
                - `2`: daily energy for the calendar month corresponding to `start_date` (`deviceRealtimeDto` is `None`)
                - `3`: (not implemented)
                - `4`: monthly energy for the calendar year corresponding to `start_date` (`deviceRealtimeDto` is `None`)
                - `5`: yearly energy, `start_date` is ignored (?) (`deviceRealtimeDto` is `None`)

        Returns:
            Statistics and metrics corresponding to the defined period. Example:
                ``` py
                {
                    "selfPowered": 0,
                    "deviceRealtimeDto": {
                        "solarPowerDps": {
                            "1732129500": 0.0,
                            "1732129800": 0.0,
                            ...
                            "1732132800": 0.0,
                        },
                        "batteryPowerDps": {...},
                        "gridPowerDps": {...},
                        "meterPowerDps": {...},
                        "homePowerDps": {...},
                        "epsPowerDps": {...},
                    },
                    "deviceStatisticsDto": {
                        "consumptionEnergy": 0.0,
                        "fromBattery": 0.0,
                        "toBattery": 0.0,
                        "fromGrid": 0.0,
                        "toGrid": 0.0,
                        "fromSolar": 0.0,
                        "eps": 0.0,
                    },
                    "insightConsumptionDataDto": {
                        "fromBatteryDps": {
                            "1733976000": 0.0,
                            "1733889600": 0.0,
                            ...
                            "1734062400": 0.0,
                        },
                        "toBatteryDps": {...},
                        "fromGridDps": {...},
                        "toGridDps": {...},
                        "fromSolarDps": {...},
                        "homeEnergyDps": {...},
                        "epsDps": {...},
                        "selfPoweredDps": {...},
                    },
                }
                ```

        """
        logger.info("Get insight for device %d", device_id)
        start_ts = int(start_date.timestamp() * 1000)  # timestamp in milliseconds
        return self._post(
            "/api/client/v2/device/three/device/insight",
            payload={
                "deviceId": device_id,
                "timestamp": start_ts,
                "periodType": period_type,
            },
        )



"""Top-level module for importing the Ecos class."""

from ._async_client import AsyncEcos
from ._client import Ecos
from ._exceptions import ApiResponseError, HttpError, InvalidJsonError

__all__ = ["Ecos", "AsyncEcos", "ApiResponseError", "HttpError", "InvalidJsonError"]



"""Unit tests for synchronous Ecos class."""

import asyncio
from datetime import datetime
import logging
import threading

from aiohttp import web
import pytest

import ecactus

from .mock_server import EcosMockServer  # noqa: TID251

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(levelname)s - %(message)s"
)

LOGIN = "test@test.com"
PASSWORD = "password"


def run_server(runner, host="127.0.0.1", port=8080):
    """Run the server."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_until_complete(runner.setup())
    site = web.TCPSite(runner, host, port)
    loop.run_until_complete(site.start())
    loop.run_forever()


@pytest.fixture(autouse=True, scope="session")
def mock_server():
    """Start a mock server in a thread and return it."""
    localhost = "127.0.0.1"
    # Find an unused localhost port from 1024-65535 and return it.
    import contextlib
    import socket

    with contextlib.closing(socket.socket(type=socket.SOCK_STREAM)) as sock:
        sock.bind((localhost, 0))
        unused_tcp_port = sock.getsockname()[1]
    server = EcosMockServer(
        host=localhost, port=unused_tcp_port, login=LOGIN, password=PASSWORD
    )
    server.setup_routes()
    runner = web.AppRunner(server.app)
    thread = threading.Thread(
        target=run_server, args=(runner, server.host, server.port)
    )
    thread.daemon = True  # daeamon thread will be killed when main thread ends
    thread.start()
    server.url = f"http://{server.host}:{server.port}"
    return server


@pytest.fixture(scope="session")
def client(mock_server):
    """Return an ECOS client."""
    return ecactus.Ecos(url=mock_server.url)


def test_login(mock_server, client):
    """Test login."""
    with pytest.raises(ecactus.ApiResponseError):
        client.login("wrong_login", "wrong_password")
    client.login(LOGIN, PASSWORD)
    assert client.access_token == mock_server.access_token
    assert client.refresh_token == mock_server.refresh_token


def test_get_user_info(caplog, mock_server, client):
    """Test get user info."""
    caplog.set_level(logging.DEBUG)
    user_info = client.get_user_info()
    assert user_info["username"] == LOGIN


def test_get_homes(mock_server, client):
    """Test get homes."""
    homes = client.get_homes()
    assert homes[1]["homeName"] == "My Home"


def test_get_devices(mock_server, client):
    """Test get devices."""
    with pytest.raises(ecactus.ApiResponseError) as err:
        client.get_devices(home_id=0)
    assert str(err.getrepr(style="value")) == "API call failed: 20450 Home does not exist."

    devices = client.get_devices(home_id=9876543210987654321)
    assert devices[0]["deviceAliasName"] == "My Device"


def test_get_all_devices(mock_server, client):
    """Test get all devices."""
    devices = client.get_all_devices()
    assert devices[0]["deviceAliasName"] == "My Device"


def test_get_realtime_device_data(mock_server, client):
    """Test get realtime device data."""
    with pytest.raises(ecactus.HttpError) as err:
        client.get_realtime_device_data(device_id=0)
    assert str(err.getrepr(style="value")) == "HTTP error: 401 unauthorized device"

    data = client.get_realtime_device_data(device_id=1234567890123456789)
    assert len(data["solarPowerDps"]) > 0


def test_get_realtime_home_data(mock_server, client):
    """Test get reatime home data."""
    with pytest.raises(ecactus.ApiResponseError) as err:
        client.get_devices(home_id=0)
    assert str(err.getrepr(style="value")) == "API call failed: 20450 Home does not exist."

    data = client.get_realtime_home_data(home_id=9876543210987654321)
    assert data.get("homePower") is not None


async def test_get_history(mock_server, client):
    """Test get history."""
    now = datetime.now()
    with pytest.raises(ecactus.HttpError) as err:
        client.get_history(device_id=0, start_date=now, period_type=0)
    assert str(err.getrepr(style="value")) == "HTTP error: 401 unauthorized device"

    with pytest.raises(ecactus.ApiResponseError) as err:
        client.get_history(device_id=1234567890123456789, start_date=now, period_type=5)
    assert str(err.getrepr(style="value")) == "API call failed: 20424 Parameter verification failed"

    data = client.get_history(device_id=1234567890123456789, start_date=now, period_type=4)
    assert len(data["homeEnergyDps"]) == 1

    #TODO other period types

async def test_get_insight(mock_server, client):
    """Test get insight."""
    now = datetime.now()
    with pytest.raises(ecactus.HttpError) as err:
        client.get_insight(device_id=0, start_date=now, period_type=0)
    assert str(err.getrepr(style="value")) == "HTTP error: 401 unauthorized device"

    with pytest.raises(ecactus.ApiResponseError) as err:
        client.get_insight(device_id=1234567890123456789, start_date=now, period_type=1)
    assert str(err.getrepr(style="value")) == "API call failed: 20404 Parameter verification failed"

    data = client.get_insight(device_id=1234567890123456789, start_date=now, period_type=0)
    assert len(data["deviceRealtimeDto"]["solarPowerDps"]) > 1


# TODO test wrong login payload
# TODO test 404
# TODO test all call not authorized
# TODO test bad method (ex GET in place of POST)



"""Base class for interacting with the ECOS API."""

import logging
from typing import Any

# Configure logging
logger = logging.getLogger(__name__)

JSON = Any


class _BaseEcos:
    """Base class for interacting with the ECOS API."""

    def __init__(
        self,
        datacenter: str | None = None,
        url: str | None = None,
        access_token: str | None = None,
        refresh_token: str | None = None,
    ) -> None:
        """Initialize a session with ECOS API.

        Args:
            datacenter (Optional[str]): The location of the ECOS API datacenter.
                Can be one of `CN`, `EU`, or `AU`. If not specified and `url` is not provided,
                a `ValueError` is raised.
            url (Optional[str]): The URL of the ECOS API. If specified, `datacenter` is ignored.
            access_token (Optional[str]): The access token for authentication with the ECOS API.
            refresh_token (Optional[str]): The refresh token for authentication with the ECOS API.

        Raises:
            ValueError: If `datacenter` is not one of `CN`, `EU`, or `AU` and `url` is not provided.

        """
        logger.info("Initializing session")
        self.access_token = access_token
        self.refresh_token = refresh_token
        # TODO: get datacenters from https://dcdn-config.weiheng-tech.com/prod/config.json
        datacenters = {
            "CN": "https://api-ecos-hu.weiheng-tech.com",
            "EU": "https://api-ecos-eu.weiheng-tech.com",
            "AU": "https://api-ecos-au.weiheng-tech.com",
        }
        if url is None:
            if datacenter is None:
                raise ValueError("url or datacenter not specified")
            if datacenter not in datacenters:
                raise ValueError(
                    "datacenter must be one of {}".format(", ".join(datacenters.keys()))
                )
            self.url = datacenters[datacenter]
        else:  # url specified, ignore datacenter
            self.url = url.rstrip("/")  # remove trailing / from url
